/*****************************************************************************/
/*
  controller.c: control strategy.
*/
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include "../useful/trajectory/trajectory.h"
#include "main.h"
#include "main2.h"

/*****************************************************************************/

/* Controller states */
#define UNKNOWN_STATE 0
#define WAITING 1
#define STOP 3

/*****************************************************************************/
/* Control variables */

DOFS desired_trajectory[N_DOFS];
DOFS current_desireds[N_DOFS];

/*****************************************************************************/
/*****************************************************************************/

void reset_score( SIM *s )
{
  s->summed_score = 0.0;
  s->discounted_score = 0.0;
  s->discount_to_power = 1.0;
  s->crashed_score = 0; /* amount of time walker is crashed. */
  s->torque_score = 0; /* sum of torque penalties */
  s->knee_score = 0; /* penalty on bending knee wrong way. */
  s->speed_score = 0;
  s->clearance_score = 0;
  s->f_x_score = 0;
  s->f_z_score = 0;
  s->pitchd_score = 0;
}

/*****************************************************************************/
/*
Could normalize torque and other penalties w.r.t. distance
Could penalize deviations from symmetry (if necessary).
*/

double one_step_cost( SIM *s )
{
}

/*****************************************************************************/

void update_score( SIM *s )
{
}

/*****************************************************************************/

double get_score( SIM *s )
{
}

/*****************************************************************************/
/* call this many times to restart a simulation */

int reinit_controller( SIM *s )
{
  int i;

  s->controller_state = WAITING;
  s->state_start_time = s->time;
  s->state_elapsed_time = 0.0;

  for( i = 0; i < N_X; i++ )
    {
      s->state_d[i] = 0;
    }

  for( i = 0; i < N_U; i++ )
    {
      s->controls[i] = 0;
      s->controls_actual[i] = 0;
      s->controls_ff[i] = 0;
    }

  for( i = 0; i < N_CTRL_INPUTS; i++ )
    {
      s->control_inputs[i] = 0;
      s->control_inputsd[i] = 0;
      s->control_inputs_int[i] = 0;
    }

  init_trajectory( desired_trajectory, N_X );
  init_trajectory( current_desireds, N_X );

  /* add desired position as knot to current desireds */
  for ( i = 0; i < N_DOFS; i++ )
    add_knot_point( current_desireds, i, QUINTIC_SPLINE, 0.0, 
		    s->state[i], s->state[i+N_DOFS], 0.0 );

  return 0;
}

/*****************************************************************************/
/* call this many times to restart a simulation */

int reinit_sim( SIM *s )
{
  int i;
  double min;

  for( i = 0; i < N_X; i++ )
    {
      s->state[i] = 0;
    }

  s->push = 0;
  s->push_location = 0;

  srand( s->rand_seed );

  s->time = 0.0;

  reinit_controller( s );
}

/*****************************************************************************/
/* Call this once to do one time operations like memory allocation */

int init_sim( SIM *s )
{
  init_dynamics( s );
  reinit_sim( s );
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

void
run_trajectory( SIM *s, double time )
{
  int i;
  double position, velocity, acceleration;

  for ( i = 0; i < N_DOFS; i++ )
    {
      if ( !lookup_spline3( desired_trajectory, i, time,
			    current_desireds, &position,
			    &velocity, &acceleration ) )
	{
	  printf( "Lookup failure.\n" );
	  exit( -1 );
	}
      /* update current desired position */
      set_first_knot_to( current_desireds, i, QUINTIC_SPLINE, time,
			 position, velocity, acceleration );
      s->state_d[i] = position;
      s->state_d[i+N_DOFS] = velocity;
    }
}

/*****************************************************************************/

void 
zero_all_integrators( SIM *s )
{
  int i;

  for( i = 0; i < N_CTRL_INPUTS; i++ )
    {
      s->control_inputs_int[i] = 0;
    }
}

/*****************************************************************************/

void
run_servos( SIM *s )
{
  int i, j;

  // First load up control_inputs: errors
  for( i = 0; i < N_DOFS; i++ )
    s->control_inputs[i] = s->state[i] - s->state_d[i];

  // derivatives
  for( i = 0, j = N_DOFS; i < N_DOFS; i++, j++ )
    s->control_inputsd[i] = s->state[j] - s->state_d[j];

  // integrals
  for( i = 0; i < N_CTRL_INPUTS; i++ )
    s->control_inputs_int[i] += s->control_inputs[i];

  for( i = 0; i < N_U; i++ )
    {
      s->controls[i] = 0;
      for( j = 0; j < N_CTRL_INPUTS; j++ )
	{
	  s->controls[i] += - s->pos_gains[i][j]*s->control_inputs[j]
	    - s->vel_gains[i][j]*s->control_inputsd[j]
	    - s->int_gains[i][j]*s->control_inputs_int[j]
	    + s->controls_ff[i];
	  /*
	  printf( "%g: %d %d; %g %g; %g %g; %g; %g\n", s->time, i, j,
		  s->control_inputs[j], 
		  s->pos_gains[i][j],
		  s->control_inputsd[j],
		  s->vel_gains[i][j],
		  s->controls_ff[i],
		  s->controls[i] );
	  */
	}
      s->controls_actual[i] = s->controls[i];
    }

  /* Perturbation */
  if ( s->time <= 0.01 )
    {
      s->push = 1.0;
      s->push_location = 1.0;
    }
  else
    {
      s->push = 0.0;
      s->push_location = 0.0;
    }
}

/*****************************************************************************/

void
run_state_machine( SIM *s )
{

  /* Debounce state transitions */
  if ( s->time - s->state_start_time < 0.001 )
    return;
  s->state_elapsed_time = s->time - s->state_start_time;

  /* Do actions and transitions */
  switch( s->controller_state )
    {
    case WAITING:
      break;
    default:
      fprintf( stderr, "Unknown state %d in run_state_machine.\n",
	       s->controller_state );
      exit( -1 );
    }
}

/*****************************************************************************/

int controller( SIM *s )
{
  int i;

  /* Kinematics already computed by SDFAST */
  /* Forces already computed by SDFAST */

  run_state_machine( s );

  run_trajectory( s, s->time );

  run_servos( s );

  update_score( s );

  return 0;
}

/*****************************************************************************/
