/************************************************************************/
/*
  dynamics.c: This is where the numerical integration and SDFAST stuff is done.
*/
/************************************************************************/
/*
In this version we are going to use the following conventions
x horizontal
y up

pole angle uses standard angle convention, where positive angle
is counter-clockwize.
Zero angle is straight up.
*/
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "main.h"
#include "main2.h"

/************************************************************************/
/* DEFINES */

#ifdef WIN32
int isnan( double x )
{ return 0; }

int isinf( double x )
{ return 0; }
#endif

/************************************************************************/
/* GLOBAL VARIABLES */

/* Simulation structure */
SIM sim;

/************************************************************************/
/************************************************************************/
/************************************************************************/

static void forward_kinematics( SIM *s )
{
  s->locations[P_POLE_TOP][XX] = 0;
  s->locations[P_POLE_TOP][YY] 
    = s->state[S_X] - s->length*sin(s->state[S_ANGLE]);
  s->locations[P_POLE_TOP][ZZ] = s->length*cos(s->state[S_ANGLE]);
}

/************************************************************************/

int set_state( SIM *s, double time, double *state )
{
  int i; 

  s->time = time;
  for ( i = 0; i < N_X; i++ )
    s->state[i] = state[i];

  s->status = OK;

  forward_kinematics( s );

  return 1;
}

/************************************************************************/
/* Initialize sdfast  */

void init_dynamics( SIM *s )
{
  s->status = OK;
  forward_kinematics( s );
}

/**********************************************************************/

void forward_dynamics( SIM *s, double *action, double *accelerations )
{
  double a, b, c, d, e, f;
  double determinant;
  double sa, ca;
  double xd;
  double ad;

  /* These dynamics were derived for straight up to be the zero position,
     so I am modifying them here so straight down is the zero position
     by subptracting PI from the shoulder angle.
  */
  sa = sin( s->state[S_ANGLE] );
  ca = cos( s->state[S_ANGLE] );

  xd = s->state[S_XD];
  ad = s->state[S_ANGLED];

  /* Flip signs on push
a1dd (m1 + m2) - a2dd*l2cm*m2*Cos(a2) =
  fx1 + fx2 - Power(a2d,2)*l2cm*m2*Sin(a2)
a2dd Ijoint - a1dd*l2cm*m2*Cos(a2) =
 -fx2*r2*Cos(a2) + G*l2cm*m2*Sin(a2)
  */

  d = s->Ijoint;
  c = -s->m2*s->l2cm*ca;
  f = -s->push*s->push_location*ca - s->viscous_friction[1]*ad
    + s->g*s->l2cm*s->m2*sa;

  b = c;
  a = s->m1 + s->m2;
  e = action[0] + s->push - s->viscous_friction[0]*xd - ad*ad*s->l2cm*s->m2*sa;
  
  determinant = a*d - b*c;
  accelerations[0] = (d*e - b*f)/determinant;
  accelerations[1] = (-c*e + a*f)/determinant;
}

/************************************************************************/
/* This is what is called on each integration step */

void integrate_one_time_step( SIM *s )
{ 
  int i;
  int is_bad_number = 0;
  double new_velocities[N_DOFS];
  double accelerations[N_DOFS];

  if( s->status == CRASHED )
    {
      s->time += s->time_step;
      return;
    }

  forward_dynamics( s, s->controls_actual, accelerations );

  // Trapezoidal integration
  for( i = 0; i < N_DOFS; i++ )
    {
      new_velocities[i] = s->state[i + N_DOFS]
	+ accelerations[i]*s->time_step;
      s->state[i] += 0.5*(new_velocities[i] 
			  + s->state[i + N_DOFS])*s->time_step;
      s->state[i + N_DOFS] = new_velocities[i];
    }
  s->time += s->time_step;

  for( i = 0; i < N_X; i++ )
    {
      if ( isnan( s->state[i] ) || isinf( s->state[i]) )
	{
	  fprintf( stderr, "Inf/Nan detected.\n" );
	  s->status = CRASHED;
	  return;
	}
    }

  forward_kinematics( s );
}

/************************************************************************/
/************************************************************************/
/************************************************************************/
/************************************************************************/
