/******************************************************************/
/* 
Defines, typedefs, and structs
*/
/******************************************************************/

/* Indices */

/* These are handy constants. XX rather than X to avoid math.h conflicts */
#define XX 0
#define YY 1
#define ZZ 2
#define XYZ 3

/* Status flags */
#define OK 0
#define CRASHED 1

/* States */
#define S_X 0
#define S_ANGLE 1
#define S_XD 2
#define S_ANGLED 3
#define N_X 4
#define N_DOFS 2
#define N_CTRL_INPUTS 2

// Controls
#define U_FORCE 0
#define N_U 1

// Points: useful for graphics, knowing where stuff is.
#define P_POLE_TOP 0
#define N_LOCATIONS 1 

/* This is used to allocate parameter vectors etc. */
#define MAX_N_PARAMETERS 100

/* This is used to allocate sdfast state */
#define MAX_N_SDFAST_STATE 100

/******************************************************************/

/* For each parameter read in from a file we have this information */
typedef struct parameter
{
  char *name;
  double value;
  int optimize;
  int regularize;
  double nominal_value;
  double regularize_weight;
  struct parameter *next;
  double *pointer;
} PARAMETER;

/******************************************************************/
/******************************************************************/
/******************************************************************/
/******************************************************************/
/******************************************************************/
/******************************************************************/
// Overall simulation structure.

typedef struct {

  // *** COMMON VARIABLES ACROSS ALL SIMULATIONS

  double time_step; // controller time step    
  double time;      // current time in simulation
  int status;       // Are we CRASHED? 

  // State
  double state[N_X];

  // Location of various body parts.
  double locations[N_LOCATIONS][XYZ];

  // commanded torques
  double controls[N_U];

  // torques (actually applied)
  double controls_actual[N_U];

  // command limits
  double controls_min[N_U];
  double controls_max[N_U];

  // *** KINEMATIC AND DYNAMIC PARAMETERS

  double length; // pole length
  double l2cm;
  double g;
  double m1;
  double m2;
  double Ijoint;
  double viscous_friction[2];

  // *** DYNAMICS **************

  /* Perturbations */
  double push;
  double push_location;

  // *** CONTROLLER **************

  // controller variables
  int controller_state;
  double state_start_time;
  double state_elapsed_time;

  // control inputs
  double control_inputs[N_CTRL_INPUTS]; // positions and angles
  double control_inputsd[N_CTRL_INPUTS]; // velocity
  double control_inputs_int[N_CTRL_INPUTS]; // integrated

  // desired positions and velocities
  double state_d[N_X];

  // gains xxx[torque][sensor]
  double pos_gains[N_U][N_CTRL_INPUTS];
  double vel_gains[N_U][N_CTRL_INPUTS];
  double int_gains[N_U][N_CTRL_INPUTS];

  // feedforwards
  double controls_ff[N_U];

  // *** PATTERN GENERATION ****************

  // *** OPTIMIZATION PARAMETERS ***************

  /* Objective function */
  double desired_speed;
  double clearance;
  double crashed_penalty_weight;
  double speed_penalty_weight;
  double torque_penalty_weight;
  double clearance_penalty_weight;
  double f_x_penalty_weight;
  double f_z_penalty_weight;
  double pitchd_penalty_weight;
  double summed_score;
  double discount;
  double discount_to_power;
  double discounted_score;
  double crashed_score;
  double torque_score;
  double knee_score;
  double speed_score;
  double clearance_score;
  double f_x_score;
  double f_z_score;
  double pitchd_score;
  double one_step_torque_score;
  double one_step_knee_score;
  double one_step_speed_score;
  double one_step_clearance_score;
  double one_step_f_x_score;
  double one_step_f_z_score;
  double one_step_pitchd_score;
  double one_step_crashed_score;
  double one_step_total_score;

  /* Useful variables for optimization */
  int n_parameters;
  double all_time_low_cost;
  double debug_criterion;
  int func_calls;
  int n_func_calls_per_eval;
  PARAMETER *params;
  char output_file[10000];
  int iter;

  // *** MISCELLANEOUS ***********************

  // simulation duration
  double duration;

  // random stuff
  // Make this thread safe: random number generator seed for rand_r() and rand()
  int rand_seed;
  double rand_scale;

  /* debugging message flags */
  int controller_print;

} SIM;

/******************************************************************/
